/**
 * @file
 * VuoFileUtilities implementation.
 *
 * @copyright Copyright © 2012–2014 Kosada Incorporated.
 * This code is proprietary, and may not be redistributed.
 * For more information, see http://vuo.org/license.
 */

#include <libgen.h>
#include <dirent.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <fstream>
#include <iostream>
#include <mach-o/dyld.h>
#include "VuoFileUtilities.hh"
#include "VuoStringUtilities.hh"


/**
 * Splits @c path into its directory, file name, and file extension.
 *
 * @param[in] path The path to split.
 * @param[out] dir The directory in @c path, if any. Ends with a file separator (e.g. '/') if there was one in @c path.
 * @param[out] file The file name in @c path, if any.
 * @param[out] extension The file extension in @c path, if any. Does not include the '.' character.
 */
void VuoFileUtilities::splitPath(string path, string &dir, string &file, string &extension)
{
	dir = "";
	file = "";
	extension = "";
	const char FILE_SEPARATOR = '/';

	size_t separatorIndex = path.rfind(FILE_SEPARATOR);
	string fileAndExtension;
	if (separatorIndex != string::npos)
	{
		dir = path.substr(0, separatorIndex + 1);
		if (separatorIndex < path.length())
			fileAndExtension = path.substr(separatorIndex + 1, path.length());
	}
	else
	{
		fileAndExtension = path;
	}

	size_t dotIndex = fileAndExtension.rfind(".");
	if (dotIndex != string::npos)
	{
		file = fileAndExtension.substr(0, dotIndex);
		extension = fileAndExtension.substr(dotIndex + 1, fileAndExtension.length());
	}
	else
	{
		file = fileAndExtension;
	}
}

/**
 * Creates a new temporary file, avoiding any name conflicts with existing files.
 * Creates the file in the specified @c directory if one is provided, or in "/tmp" otherwise.
 *
 * Returns the path of the file.
 */
string VuoFileUtilities::makeTmpFile(string file, string extension, string directory)
{
	if ((directory.length() > 0) && (directory.at(directory.length()-1) != '/'))
		directory += "/";
	string suffix = (extension.empty() ? "" : ("." + extension));
	string pathTemplate = directory + file + "-XXXXXX" + suffix;
	char *path = (char *)calloc(pathTemplate.length() + 1, sizeof(char));
	strncpy(path, pathTemplate.c_str(), pathTemplate.length());
	int fd = mkstemps(path, suffix.length());
	close(fd);  /// @todo Keep file open and return file handle. (https://b33p.net/kosada/node/4987)
	string pathStr(path);
	free(path);
	return pathStr;
}

/**
 * Creates a new temporary directory, avoiding any name conflicts with existing files.
 *
 * Returns the path of the directory.
 */
string VuoFileUtilities::makeTmpDir(string dir)
{
	string pathTemplate = "/tmp/" + dir + ".XXXXXX";
	char *path = (char *)calloc(pathTemplate.length() + 1, sizeof(char));
	strncpy(path, pathTemplate.c_str(), pathTemplate.length());
	mkdtemp(path);
	string pathStr(path);
	free(path);
	return pathStr;
}

/**
 * Returns the path of the default temporary directory.
 */
string VuoFileUtilities::getTmpDir(void)
{
	return "/tmp";
}

/**
 * Returns the path of Vuo.framework, or an empty string if Vuo.framework cannot be located.
 */
string VuoFileUtilities::getVuoFrameworkPath(void)
{
	// First check whether Vuo.framework is in the list of loaded dynamic libraries.
	const char *frameworkDylibRelativePath = "Vuo.framework/Versions/" VUO_VERSION_STRING "/Vuo";
	for(unsigned int i=0; i<_dyld_image_count(); ++i)
	{
		const char *dylibPath = _dyld_get_image_name(i);
		if (VuoStringUtilities::endsWith(dylibPath, frameworkDylibRelativePath))
		{
			string path = dylibPath;
			string dir, file, ext;
			splitPath(path, dir, file, ext);		// remove "Vuo"
			path = dir.substr(0, dir.length() - 1);	// remove "/"
			splitPath(path, dir, file, ext);		// remove VUO_VERSION_STRING
			path = dir.substr(0, dir.length() - 1);	// remove "/"
			splitPath(path, dir, file, ext);		// remove "Versions"
			path = dir.substr(0, dir.length() - 1);	// remove "/"
			if (fileExists(path))
				return path;
		}
	}

	// Failing that, check for a Vuo.framework bundled with the executable app.
	char executablePath[PATH_MAX + 1];
	uint32_t size = sizeof(executablePath);

	if (! _NSGetExecutablePath(executablePath, &size))
	{
		char cleanedExecutablePath[PATH_MAX + 1];

		realpath(executablePath, cleanedExecutablePath);  // remove extra references (e.g., "/./")
		string path = cleanedExecutablePath;
		string dir, file, ext;
		splitPath(path, dir, file, ext);		// remove executable name
		path = dir.substr(0, dir.length() - 1);	// remove "/"
		splitPath(path, dir, file, ext);		// remove "MacOS"
		path = dir.substr(0, dir.length() - 1);	// remove "/"
		path += "/Frameworks/Vuo.framework";

		if (fileExists(path))
			return path;
	}

	// Failing that, check for ~/Library/Frameworks/Vuo.framework.
	string userFrameworkPath = string(getenv("HOME")) + "/Library/Frameworks/Vuo.framework";
	if (fileExists(userFrameworkPath))
		return userFrameworkPath;

	// Failing that, check for /Library/Frameworks/Vuo.framework.
	string systemFrameworkPath = "/Library/Frameworks/Vuo.framework";
	if (fileExists(systemFrameworkPath))
		return systemFrameworkPath;

	// Give up.
	return "";
}

/**
 * Returns the first index at which content can be inserted into a string that was read from a file.
 * This comes after the Unicode BOM, if present.
 */
size_t VuoFileUtilities::getFirstInsertionIndex(string s)
{
	char bomUtf8[] = { 0xEF, 0xBB, 0xBF };
	char bomUtf16Be[] = { 0xFE, 0xFF };
	char bomUtf16Le[] = { 0xFF, 0xFE };
	string boms[] = { bomUtf8, bomUtf16Be, bomUtf16Le };
	for (int i = 0; i < 3; ++i)
	{
		bool isMatch = true;
		for (int j = 0; j < boms[i].length(); ++j)
			if (boms[i][j] != s[j])
				isMatch = false;

		if (isMatch)
			return boms[i].length();
	}
	return 0;
}

/**
 * Reads from standard input into a string until the end-of-file is reached.
 */
string VuoFileUtilities::readStdinToString(void)
{
	// http://stackoverflow.com/questions/201992/how-to-read-until-eof-from-cin-in-c

	string contents;
	string line;
	while (getline(cin, line))
	   contents += line;

	return contents;
}

/**
 * Reads the whole contents of the file into a string.
 */
string VuoFileUtilities::readFileToString(string path)
{
	// http://stackoverflow.com/questions/2602013/read-whole-ascii-file-into-c-stdstring

	ifstream in(path.c_str(), ios::in | ios::binary);
	if (! in)
	{
		VLog("Error: Couldn't read file at path '%s'.", path.c_str());
		return "";
	}

	string contents;
	in.seekg(0, ios::end);
	contents.resize(in.tellg());
	in.seekg(0, ios::beg);
	in.read(&contents[0], contents.size());
	in.close();
	return contents;
}

/**
 * Writes the array of bytes to the file.
 *
 * If the file already exists, it gets overwritten.
 */
void VuoFileUtilities::writeRawDataToFile(const char *data, size_t numBytes, string file)
{
	FILE *f = fopen(file.c_str(), "wb");
	fwrite(data, sizeof(char), numBytes, f);
	fclose(f);
}

/**
 * Writes the string to the file.
 *
 * If the file already exists, it gets overwritten.
 */
void VuoFileUtilities::writeStringToFile(string s, string file)
{
	writeRawDataToFile(s.data(), s.length(), file);
}

/**
 * Returns true if the file exists.
 */
bool VuoFileUtilities::fileExists(string path)
{
	return access(path.c_str(), 0) == 0;
}

/**
 * Searches a directory for files.
 *
 * @param dirPath The directory to search in. Only the top level is searched.
 * @param archiveExtensions The file extensions for archives to search in. Any archive with one of these extensions
 *		found in the top level of the directory will be searched recursively.
 * @param shouldSearchRecursively If true, the directory will be searched searched recursively.
 * @return All files found.
 */
set<VuoFileUtilities::File *> VuoFileUtilities::findAllFilesInDirectory(string dirPath, set<string> archiveExtensions,
																		bool shouldSearchRecursively)
{
	set<File *> files;

	DIR *d = opendir(dirPath.c_str());
	if (! d)
	{
		if (access(dirPath.c_str(), F_OK) != -1)
			VLog("Error: Couldn't open directory '%s' to find files in it.", dirPath.c_str());
		return files;
	}

	struct dirent *de;
	while( (de=readdir(d)) )
	{
		string fileName = de->d_name;
		string relativeFilePath = dirPath + "/" + fileName;

		if (fileName == "." || fileName == "..")
			continue;

		bool isArchive = false;
		for (set<string>::iterator archiveExtension = archiveExtensions.begin(); archiveExtension != archiveExtensions.end(); ++archiveExtension)
			if (VuoStringUtilities::endsWith(fileName, "." + *archiveExtension))
				isArchive = true;

		if (isArchive)
		{
			set<File *> fs = findAllFilesInArchive(relativeFilePath);
			if (fs.empty())
				isArchive = false;
			else
				files.insert(fs.begin(), fs.end());
		}

		if (! isArchive)
		{
			bool shouldSearchDir = false;
			if (shouldSearchRecursively)
			{
				struct stat st_buf;
				int status = lstat(relativeFilePath.c_str(), &st_buf);  // Unlike stat, lstat doesn't follow symlinks.
				if (! status && S_ISDIR(st_buf.st_mode))
					shouldSearchDir = true;
			}

			if (shouldSearchDir)
			{
				set<File *> filesInDir = findAllFilesInDirectory(relativeFilePath, archiveExtensions, true);
				for (set<File *>::iterator i = filesInDir.begin(); i != filesInDir.end(); ++i)
				{
					File *f = *i;
					f->dirPath = dirPath;
					f->filePath = fileName + "/" + f->filePath;
				}
				files.insert(filesInDir.begin(), filesInDir.end());
			}
			else
			{
				File *f = new File(dirPath, fileName);
				files.insert(f);
			}
		}
	}

	closedir(d);

	return files;
}

/**
 * Searches a directory for files with a given extension.
 *
 * @param dirPath The directory to search in. Only the top level is searched.
 * @param archiveExtensions The file extensions for archives to search in. Any archive with one of these extensions
 *		found in the top level of the directory will be searched recursively.
 * @param extensions The file extensions to search for, without the '.' character (e.g. "bc" not ".bc").
 * @return All files found.
 */
set<VuoFileUtilities::File *> VuoFileUtilities::findFilesInDirectory(string dirPath, set<string> extensions, set<string> archiveExtensions)
{
	set<File *> allFiles = findAllFilesInDirectory(dirPath, archiveExtensions);
	set<File *> matchingFiles;

	for (set<File *>::iterator i = allFiles.begin(); i != allFiles.end(); ++i)
	{
		bool endsWithExtension = false;
		for (set<string>::iterator extension = extensions.begin(); extension != extensions.end(); ++extension)
			if (VuoStringUtilities::endsWith((*i)->getRelativePath(), "." + *extension))
				endsWithExtension = true;

		if (endsWithExtension)
			matchingFiles.insert(*i);
		else
			delete *i;
	}

	return matchingFiles;
}

/**
 * Recursively searches an archive for files.
 *
 * @param archivePath The archive to search in.
 * @return All files found.
 */
set<VuoFileUtilities::File *> VuoFileUtilities::findAllFilesInArchive(string archivePath)
{
	set<File *> files;

	Archive *archive = new Archive(archivePath);
	if (! archive->zipArchive)
	{
		delete archive;
		return files;
	}

	mz_uint numFiles = mz_zip_reader_get_num_files(archive->zipArchive);
	for (mz_uint i = 0; i < numFiles; ++i)
	{
		mz_zip_archive_file_stat file_stat;
		bool success = mz_zip_reader_file_stat(archive->zipArchive, i, &file_stat);
		if (! success)
		{
			VLog("Error: Couldn't read file '%u' in zip archive '%s'.", i, archive->path.c_str());
			break;
		}

		File *f = new File(archive, file_stat.m_filename);
		files.insert(f);
	}

	if (archive->referenceCount == 0)
		delete archive;

	return files;
}

/**
 * Recursively searches an archive for files within a given directory and with a given extension.
 *
 * @param archivePath The archive to search in.
 * @param dirPath The directory to search in within the archive. The path should be relative to the archive's root.
 *			The path should omit the trailing file separator (e.g. "examples" not "examples/").
 * @param extensions The file extensions to search for, without the '.' character (e.g. "bc" not ".bc").
 * @return All files found.
 */
set<VuoFileUtilities::File *> VuoFileUtilities::findFilesInArchive(string archivePath, string dirPath, set<string> extensions)
{
	set<File *> allFiles = findAllFilesInArchive(archivePath);
	set<File *> matchingFiles;

	for (set<File *>::iterator i = allFiles.begin(); i != allFiles.end(); ++i)
	{
		File *f = *i;

		string dir, file, ext;
		splitPath(f->getRelativePath(), dir, file, ext);
		bool beginsWithDir = VuoStringUtilities::beginsWith(dir, dirPath);

		bool endsWithExtension = false;
		for (set<string>::iterator extension = extensions.begin(); extension != extensions.end(); ++extension)
			if (VuoStringUtilities::endsWith(f->getRelativePath(), "." + *extension))
				endsWithExtension = true;

		if (beginsWithDir && endsWithExtension)
			matchingFiles.insert(f);
		else
			delete f;
	}

	return matchingFiles;
}

/**
 * Returns the contents of a file within an archive.
 *
 * If the file doesn't exist, returns an empty string.
 *
 * @param archivePath The archive containing the file.
 * @param filePath The file. The path should be relative to the archive's root.
 * @return The file's contents.
 */
string VuoFileUtilities::getArchiveFileContentsAsString(string archivePath, string filePath)
{
	string contents;
	set<File *> archiveFiles = findAllFilesInArchive(archivePath);
	for (set<File *>::iterator i = archiveFiles.begin(); i != archiveFiles.end(); ++i)
	{
		File *file = *i;
		if (filePath == file->getRelativePath())
			contents = file->getContentsAsString();
		delete file;
	}

	return contents;
}

/**
 * Creates an archive with an open zip archive handle.
 *
 * If the file can't be opened as an archive, the zip archive handle is null.
 */
VuoFileUtilities::Archive::Archive(string path)
{
	this->path = path;
	this->referenceCount = 0;

	// mz_zip_reader_init_file sometimes ends up with a garbage function pointer if the mz_zip_archive isn't initialized to zeroes
	this->zipArchive = (mz_zip_archive *)calloc(1, sizeof(mz_zip_archive));

	bool success = mz_zip_reader_init_file(zipArchive, path.c_str(), 0);
	if (! success)
	{
		free(zipArchive);
		zipArchive = NULL;
	}
}

/**
 * Closes the zip archive handle.
 */
VuoFileUtilities::Archive::~Archive(void)
{
	if (zipArchive)
		mz_zip_reader_end(zipArchive);
}

/**
 * Creates a reference to a file that is in a directory.
 */
VuoFileUtilities::File::File(string dirPath, string filePath)
{
	this->filePath = filePath;
	this->dirPath = dirPath;
	this->archive = NULL;
}

/**
 * Creates a reference to a file that is in an archive.
 *
 * After this constructor is called, it's important to call the destructor.
 * The archive will be closed when all File objects that used it are destroyed.
 */
VuoFileUtilities::File::File(Archive *archive, string filePath)
{
	this->filePath = filePath;
	this->archive = archive;
	++archive->referenceCount;
}

/**
 * If this file is in an archive and it was the last File object using the Archive object,
 * closes the archive.
 */
VuoFileUtilities::File::~File(void)
{
	if (archive && --archive->referenceCount == 0)
		delete archive;
}

/**
 * Returns true if this file is inside an archive.
 */
bool VuoFileUtilities::File::isInArchive(void)
{
	return (archive != NULL);
}

/**
 * Returns the path of the archive that contains this file.
 *
 * If this file is not inside an archive, returns an empty string.
 */
string VuoFileUtilities::File::getArchivePath(void)
{
	return (archive ? archive->path : "");
}

/**
 * Returns the path of the file, relative to its enclosing directory or archive.
 */
string VuoFileUtilities::File::getRelativePath(void)
{
	return filePath;
}

/**
 * Returns the contents of the file as an array of bytes.
 *
 * @param[out] numBytes The size of the returned array.
 * @return The array of bytes. The caller is responsible for freeing it.
 */
char * VuoFileUtilities::File::getContentsAsRawData(size_t &numBytes)
{
	char * buffer;
	numBytes = 0;

	if (! archive)
	{
		// http://www.cplusplus.com/reference/cstdio/fread/

		string fullPath = (dirPath.empty() ? "" : (dirPath + "/")) + filePath;

		FILE *pFile = fopen ( fullPath.c_str() , "rb" );
		if (pFile==NULL)
		{
			VLog("Error: Couldn't open file '%s' for reading.", fullPath.c_str());
			return NULL;
		}

		// obtain file size:
		fseek (pFile , 0 , SEEK_END);
		size_t lSize = ftell (pFile);
		rewind (pFile);

		// allocate memory to contain the whole file:
		buffer = (char*) malloc (sizeof(char)*lSize);

		// copy the file into the buffer:
		numBytes = fread (buffer,1,lSize,pFile);
		fclose(pFile);
		if (numBytes != lSize)
		{
			VLog("Error: Couldn't read file '%s'.", fullPath.c_str());
			free(buffer);
			return NULL;
		}
	}
	else
	{
		buffer = (char *)mz_zip_reader_extract_file_to_heap(archive->zipArchive, filePath.c_str(), &numBytes, 0);
	}

	return buffer;
}

/**
 * Returns the contents of the file as a string.
 */
string VuoFileUtilities::File::getContentsAsString(void)
{
	size_t numBytes;
	char *buffer = getContentsAsRawData(numBytes);
	string s(buffer, numBytes);
	free(buffer);
	return s;
}
